!
! hblanc --- band Lanczos for complex Hermitian matrices
! Hans-Dieter Meyer
!
!
! Copyright (c) 2003 Hans-Dieter Meyer
! Copyright (c) 2003 Thomas Sommerfeld
! Copyright (c) 2004-2006 Christian Buth
!
! This library is free software; you can redistribute it and/or
! modify it under the terms of the GNU Lesser General Public
! License as published by the Free Software Foundation; either
! version 2.1 of the License, or (at your option) any later version.
!
! This library is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! Lesser General Public License for more details.
!
! You should have received a copy of the GNU Lesser General Public
! License along with this library; if not, write to the Free Software
! Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
!
!
!     ----------------------------------------------------------------
!
!     BAND LANCZOS for complex Hermitian matrices
!
!     large parts of this routine are taken from the block lanczos
!     program by H-D. Meyer and U. V. Riss
!
!     a matrix A is transformed to band-diagonal form T
!     A is only refered to via the subroutine hmatblk, which
!     has to perform the matrix times vector-block operation
!
!     ndim     dimension of the matrix A
!     np       block-size for the iteration --> band width of T
!     niter    no. of iterations (matrix*vector operations)
!     neigenv  No of eigenvectors to be computer, zero for no
!              second eigenvector run
!     ncartsv  If > 0, Lanczos vectors are orthogonalized to the
!              cartesian starting vectors, i.e. it is assured
!              that the element of the Lanczos vectors vanishes
!              where a cartesian start vector has its only non-zero
!              entry which is unity
!     iparam   array of additional parameters
!     iparam(1)  jstart     1st iteration  (can be > 1 in restarts)
!     iparam(2)  jend       index of last iteration (output)
!     iparam(3)  neigenv    abs of no. of eigenvectors to be computed
!                neigenv =  0  : normaler Lanczos
!                neigenv = -1  : Vorlauf fr Eigenvektoren
!                neigenv >  0  : Eigenvektorlauf
!     iparam(4)  restart
!     iparam(5)  wrtrsf    write a restart file
!     iparam(6)  wrttmat   write T to a file
!     iparam(7)  ncartsv   no. of cartesian start-vectors
!                if the start-vectors are cartesian this is used
!                to normalize the lanczos-vectors on the start-block
!
!     t        contains on output the computed band matrix
!     p        contains 3*np vectors needed in the iteration
!              on input p(1,i,2) is used as the i-th start-vector
!              (it's not my fault)
!
!     z and ev are only referenced, if iparam(3) > 0
!     z        contains some eigenvectors of T from a previous run
!     ev       contains on output the corresponding eigenvectors of A
!
!     icartsv  contains the indices of the cartesian start-vectors
!
!     external hmatblk (ndim, np, pold, pnew, work, jdiesel)
!     is the mat times vector routine
!     work contains space for zmatvec and is not referenced in the lanczos
!
!     THE SUBROUTINE INDEXL COMPUTES THE CURREND INDEX OF THE LANCZOS-
!                           VECTOR CONTAINED IN THE ARRAY P.
!
!     -----------------------------------------------------------------

      subroutine hblanc(ndim, np, niter, neigenv, ncartsv, iparam, t,   &
     &     p, z, ev, icartsv, am) !!, hmatblk, mtxwork, jdiesel)
      use fortrantl
      use coadcfile
      use coadcdiag
      implicit none

      integer ndim, np, niter, neigenv, ncartsv  !!, jdiesel
      integer iparam(10), icartsv(ncartsv)
      complex(dblprec) p(ndim,np,3), t(np+1,niter)
      complex(dblprec) z(niter,neigenv), ev(ndim,neigenv) !!, mtxwork(*)
      !!!!external hmatblk
      type(tcoadcmatrix), intent(inout) :: am

      integer jstart, jend
      integer restart, wrtrsf, wrttmat
      integer np1, np3

      real(dblprec) r
      complex(dblprec) y
      integer i, j, k, nrun, ndone
      integer jj, ii, jn, in, jo, io, jjc, iic, im, ip, jp
      integer jiter, jko
      logical da

      complex(dblprec) zdotc
      external zaxpy, zdotc, zdscal

#include "iosetting.inc"

!     -------------------------------------------------------------
!     preliminaries
!     -------------------------------------------------------------
      write (6,100)
 100  format (//,2X,50('-')/,2X,'Band-Lanczos (complex '                &
     &        'Hermitian)'/)
      jstart  = iparam(1)
      jend    = iparam(2)
      restart = iparam(4)
      wrtrsf  = iparam(5)
      wrttmat = iparam(6)
      np1 = np + 1
      np3 = 3 * np
      jstart = 1
!     ---------------------------------------------------------
!     read the restartfile
!     --------------------
      if (restart .ne. 0) then
         write (6,'(4x,A)') 'Restart from file'
         inquire(file='RESTART', exist=da)
         if (.not. da) then
            write(6,*) 'RESTART File is missing.'
            stop
         end if
         open (ifile,file='RESTART',access='SEQUENTIAL',                &
     &        form='UNFORMATTED')
         read(ifile) jend, j
         if (j .ne. ndim) then
            write(6,*) 'Inconsistancy:'
            write(6,*) 'Ndim(Inputfiles)  = ', ndim
            write(6,*) 'Ndim(Restartfile) = ', j
            stop
         end if
         write(6,'(6x,A,I5)') 'Iterations done so far  ', jend
         write(6,'(6x,A,I5)') 'Iterations to be done   ', niter-jend
         if (jend .ge. niter) then
            stop 'Nothing to do'
         end if
         read(ifile) ((t(i,j), i=1, np1),  j=1, jend)
         read(ifile) ((p(i,j,1), i=1, ndim), j=1, np),                  &
     &            ((p(i,j,2), i=1, ndim), j=1, np)
         read(ifile) ncartsv
         read(ifile) (icartsv(i), i=1, ncartsv)
         close (ifile)
         jstart = jend + 1
      end if
!     ----------------------------
!     print Iterations and runtype
!     ----------------------------
      write (6,'(4x,A,I5,A,I5)') 'Iterations from ',jstart,' to ',niter
      if (neigenv .gt. 0) then
         write(6,'(4X,A,I4,A)') 'This is a second run computing ',      &
     &        neigenv,' eigenvectors'
      end if
      write(6,*)
!     ---------------------------------------------------------------
!     prepare for eigenvector run
!     ---------------------------
      if (neigenv .gt. 0) then
!     --------------------------------------
!     read the ZMAT file to z(niter,neigenv)
!     this file is created by banddia after
!     a first lanczos requiring eigenvectors
!     --------------------------------------
        inquire(file='ZMAT', exist=da)
         if (.not. da) then
            write(6,*) '**** The File ZMAT does not exist.'
            stop
         end if
         open (ifile,file='ZMAT',access='SEQUENTIAL',                   &
     &        form='UNFORMATTED')
         read(ifile) i, j
         if (i .ne. niter .or. j .ne. neigenv) then
            write(6,*) '***************   WARNING  *****************'
            write(6,*) ' ZMAT file _may_ be corrupted.'
            write(6,*) ' niter, neigenv (Args) = ', niter, neigenv
            write(6,*) ' zdim,  neigenv (File) = ', i, j
            write(6,*) ' (niter must be too small in a restart set up)'
            write(6,*)
            if (j .lt. neigenv .or. i .lt. niter) then
               stop
            end if
         end if
         do j = 1, neigenv
            read(ifile) (z(i,j), i = 1, niter)
         end do
         close (ifile)
!     -------------------------------
!     initialize or read eigenvectors
!     -------------------------------
         if (restart .eq. 0) then
            do i = 1, neigenv
               do j = 1, ndim
                  ev(j,i) = (0.0d0, 0.0d0)
               end do
            end do
            do j = 1, np
               jiter = jstart - 1 + j
               do i = 1, neigenv
                  do jko = 1, ndim
                     ev(jko,i) = ev(jko,i)+p(jko,j,2)*z(jiter,i)
                  end do
               end do
            end do

         else
!     ------------------------------------------------------
!     restart = 1
!     read eigenvectors from the EIGENV file of the last run
!     ------------------------------------------------------
            inquire(file='EIGENV', exist=da)
            if (.not. da) then
               write(6,*) 'EIGENV File is missing.'
               stop
            end if
            open (ifile,file='EIGENV',access='SEQUENTIAL',form          &
     &           ='UNFORMATTED')
            read(ifile) i, j
            if (i .ne. neigenv .or. j .ne. ndim) then
               write(6,*) 'EIGENV file may be corrupted.'
               write(6,*) 'ndim, neigenv (Args) = ', ndim, neigenv
               write(6,*) 'ndim, neigenv (File) = ', j, i
               stop
            end if
            read(ifile) i
            if (i .ne. 2) then
               write(6,*) 'itype = ',i,' ?????'
               stop
            end if
            do j = 1, neigenv
               read(ifile) (ev(i,j), i = 1, ndim)
            end do
            close(ifile)
!     ------------------------------------------
!     contributions of the restart vectors which
!     have NOT been added in the previous run!
!     ------------------------------------------
            do j = 1, np
               jiter = jstart - 1 + j
               do i = 1, neigenv
                  do jko = 1, ndim
                     ev(jko,i) = ev(jko,i) + p(jko,j,2) * z(jiter,i)
                  end do
               end do
            end do
         end if
      end if
!     ============================================================
!     end of eigenvector preparations, now start the
!
!     Main Loop of the Lanczos Iteration
!     ----------------------------------
      nrun = niter - jstart + 1
      ndone = 10

      ! Do the respective number of band Lanczos iterations
      do JJ = JSTART, NITER

         ! in -- index of block Lanczos vector of the next iteration
         ! jn -- remaining Lanczos vectors of the block to be processed
         ! cyclic(in+1) -- index of the current block Lanczos vector
         CALL INDEXL(JJ-JSTART+NP, NP, jn, IN )
!
!     when neccessary do the next matrix times block operation
!
         if (MOD(JJ-JSTART,NP)  .EQ. 0) then
            ! io -- index of the result of the block matrix times matrix operation
            io = mod(in + 1, 3) + 1
        !    write (*,*) 'Block matrix times vector product No. ', JJ
        !    write (*,*) 'in: ', in, ' io: ', io
!           call hmatblk(ndim,np,p(1,1,io),p(1,1,in),mtxwork,jdiesel)
            call bandmatvec(am, np, p(1:ndim, 1:np, io),                &
     &                              p(1:ndim, 1:np, in))
         endif
!
!     Band-Lanczos algorithm as implemented by H-D. Meyer
!
         ! The lower triangle of the band matrix is computed and stored
         ! in packed storage format
         ! Compute intermediate vectors r in Eq. (3.5a)
         do II = max(JJ-NP,1), JJ-1
            ! io -- previous block Lanczos vector
            ! jo -- previous Lanczos vector in the block
            CALL INDEXL(II-JSTART+NP3, NP, JO, IO )
            ! T(JJ, II) in packed storage (lower triangle)
            JJC = JJ - II + 1
!            do K = 1, NDIM
!               P(k,JN,IN) = P(k,JN,IN) - P(k,JO,IO)*T(JJC,II)
!            end do
            call zaxpy(ndim,-t(jjc,ii),p(1,jo,io),1,p(1,jn,in),1)
         enddo

         do II = JJ, JJ + NP - 1
            CALL INDEXL(II-JSTART, NP, JO, IO )
!            Y = (0.0d0, 0.0d0)
!            do K = 1, NDIM
!               Y = Y + conjg(P(k,JO,IO)) * P(k,JN,IN)
!            end do
            ! Compute banddiagonal matrix elements in Eq. (3.5b)
            y = zdotc(ndim,p(1,jo,io),1,p(1,jn,in),1)
            ! T(II, JJ) = Y in packed storage (lower triangle)
            IIC = II - JJ + 1
            ! CB T(IIC, JJ) = Y
            T(IIC, JJ) = conjg(Y)
!            do K = 1, NDIM
!               P(k,JN,IN) = P(k,JN,IN) - P(k,JO,IO)*Y
!            end do
            ! Compute new vector r in Eq. (3.5c)
            call zaxpy(ndim,-y,p(1,jo,io),1,p(1,jn,in),1)
         end do
!     -------------------------------------------------
!     orthogonalization wrt cartesian start-vectors
!     Vorsicht bei restart; icartsv muss definiert sein
!     -------------------------------------------------
         if (ncartsv .gt. 0) then
            do k = 1, np
               p(icartsv(k),jn,in) = (0.0d0, 0.0d0)
            end do
         end if
!     --------------------
!     normalize new vector
!     --------------------
!         r = 0.0d0
!         do K = 1, NDIM
!            r = r + |P(k,JN,IN)|**2
!         end do
         ! Compute last matrix element of the banddiagonal matrix in Eq. (3.5d)
         r = real(zdotc(ndim,p(1,jn,in),1,p(1,jn,in),1), dblprec)
         r = sqrt(r)
         ! Packed storage lower triangle
         ! T(JJ, JJ + ndim) = Tpacked(JJ + ndim - JJ + 1, JJ)
         T(NP1,JJ) = conjg(cmplx(r, 0.0D0, dblprec))
         r = 1.0d0 / r
!         do K = 1, NDIM
!            P(k,JN,IN) = P(k,JN,IN) * r
!         end do
         ! Scale vector r in Eq. (3.5e)
         ! Block Lanczos vector for the next iteration is in block in
         ! vector number jn
         call zdscal(ndim, r, p(1,jn,in), 1)
!     -------------------------------------------------------
!     sum contributions of this iteration to the eigenvectors
!     -------------------------------------------------------
         if (neigenv .gt. 0) then
            jiter = np + jj
!
!     in the eigenvector calculation the very last np iterations
!     do not contribute to the eigenvectors;
!     but in case of a later restart these iterations have to be done
!
            if (jiter .le. niter) then
               do i = 1, neigenv
                  do jko = 1, ndim
                     ev(jko,i)=ev(jko,i)+z(jiter,i)*p(jko,JN,IN)
                  end do
               end do
            end if
         end if
!
!     check how much has been done
!
         if ((jj-jstart+1)*100 / nrun .ge. ndone) then
            write(6,'(6X,I3,A)') ndone, '%'
            call flush(6)
            ndone = ndone + 10
         end if

      end do
!     ------------------------------------------------------------
!     End of the Block-Lanczos iteration
!     jend was used when the main-loop was left because of
!     a timelimit (IMB 30/90 version); this has been thrown out
!     ------------------------------------------------------------
      write(6,999)
 999  FORMAT(/,2X,'Lanczos finished',/,2X,50('-'))
      jend = niter
!     ---------------------------------------------------
!     write the bandmatrix T(np1, niter) to the file TMAT
!     for convergence and other checks
!     ---------------------------------------------------
      if (wrttmat .eq. 1) then
         open(unit=ifile,file='TMAT',access='SEQUENTIAL',               &
     &        form='UNFORMATTED')
         write (ifile) jend, np, niter
         write (ifile) t
         close (ifile)
      end if
!     -----------------------------------------------
!     write the restart file
!     this part was written by Dieter;
!     i do not really understand it, however it works
!     -----------------------------------------------
      if (wrtrsf .eq. 1) then
         open (ifile,file='RESTART',access='SEQUENTIAL',                &
     &        form='UNFORMATTED')
         CALL INDEXL(JEND-JSTART+NP, NP, JP, IP )
         IO  = MOD(IP+1,3) + 1
         IM  = MOD(IP  ,3) + 1
         write(ifile) jend, ndim
         write(ifile) ((T(I,J), I=1,NP1) ,J=1,JEND)
         write(ifile) ((P(i,j,IM), i=1,NDIM), j=JP+1,NP),               &
     &             ((P(i,j,IO), i=1,NDIM), j=1,   NP),                  &
     &             ((P(i,j,IP), i=1,NDIM), j=1,   JP)
         write(ifile) ncartsv
         write(ifile) (icartsv(i), i=1, ncartsv)
         close (ifile)
      end if


      contains

!
!     one of Dieter's more serious crimes:
!     jj -- (in) current band Lanczos iteration
!     np -- (in) number of Lanczos vectors, the size of the main space
!     i  -- (out) current index of the Lanczos block vector
!     j  -- (out) current Lanczos vector in the block vector to be processed
!
!
      subroutine indexl( jj, np, j, i )
      implicit none

      integer, intent(in) :: jj, np
      integer, intent(out) :: i, j
      integer kk

!
!     AH1:LANCZ.INDEXL
!     THIS ROUTINE COMPUTES THE INDICES J,I OF THE JJ-TH LANCZOS-VECTOR.
!
      kk = jj / np                   ! Number of block Lanczos vector times matrix
                                     ! operations that have been processed so far
      i  = mod(kk+1, 3) + 1
      j  = jj + 1 - kk * np
      end subroutine indexl
      end subroutine hblanc
